/**
 * A class that implements the MyMap interface using a hash map/table
 * data structure.
 *
 * @author <your name here>
 * @version 1.0
 */
public class MyHashMap<K,V> implements MyMap<K,V> {
    // number of items in the hash table
    private int nentries = 0;

    // initial number of buckets
    private int nbuckets = 10;

    // the threshold above which the hashtable will be resized
    static final double RESIZE_THRESHOLD = 0.75; // package private to allow testing
 
    // the actual hash table which is an array of linked lists
    // each linked list is a list of key value pairs
    // I was missing a few things with the statement I put on the board
    private MyLinkedList<KeyValuePair<K,V>>[] table;
    
    public MyHashMap() {
        
        table = new MyLinkedList[nbuckets];
        for (int i = 0; i < table.length; i++) {
           table[i] = new MyLinkedList<KeyValuePair<K,V>>();
        }        
    }

    /**
     * Key-Value Pair Container Class to hold our keys and values
     */
    private class KeyValuePair<K,V> {
        //generic types here
        K key;
        V value; 
        //constructor
        public KeyValuePair(K k, V v) {
            key = k;
            value = v;
        }
        
        /**
         * see if keys match (we don't care about the values matching)
         */
        @Override
        public boolean equals (Object o) {
           boolean eq = false;
           if (o instanceof MyHashMap.KeyValuePair) {
               KeyValuePair kvp = (KeyValuePair) o;
               eq = this.key.equals(kvp.key);
            }
            return eq;
           }
           
        
    }

    /**
     * create our own hash method
     * Math.abs is not strictly necessary, more a defensive coding move.
     */

    int hash (K key) {
        return Math.abs(key.hashCode()) % nbuckets;
        
    }
    /**
     * Associates a value with a key.
     * If adding this new (key, value) pair makes the load factor cross
     * the resize threshold, the number of buckets should be doubled and
     * all entries rehashed.
     *
     * @param k key
     * @param v value
     * @return if key was already associated with a value return
     * it - the old value I mean... otherwise return null.
     */
    public V put(K key, V value) {
         // hash key to find target bucket index
        int bindex = hash(key);
        MyLinkedList<KeyValuePair<K,V>> blist = this.table[bindex];
        
        // go over the (key, value) pairs in the bucket list (blist), looking for the key
        MyIterator<KeyValuePair<K,V>> itr = blist.iterator();
        while(itr.hasNext()){ // iterate while there are more items to visit
            KeyValuePair kvp = itr.next(); // retrieve an unvisited item            

            if (kvp.key.equals(key)){ // found the key!
                V oldValue = (V) kvp.value; // save old value                
                kvp.value = value; // update the value
                return oldValue; // return old value
            } 
        }
         
        // if we got this far, it means we're adding a new key otherwise we already returned the old value
        KeyValuePair newKvp = new KeyValuePair(key, value);
        blist.add(0, newKvp); // add at head of list
        this.nentries++; // one more entry  
        
        
        
        /**
         * ok this code in put works, as far as it goes... 
         * However, we still need to rework our hash table if things get a little too crowded.
         * still TODO: use the loadFactor() method below to determine we need to resize and do so.
         * In order to make this method easier to read, you can also create a resize() method
         * if you prefer.
         */
        
        
        
        return null;
    }

    /**
     *  Retrieves a value associated with a key. 
     *  @param k key
     *  @return the value associated with that key, null if 
     *  key not present in the map.
     */
    public V get(K key) {
        // TODO implement
        return null;
    }

    /**
     * Removes mapping associated with that key.
     * @param k key
     * @return the value associated with that key, or null if none
     */
    public V remove(K key) {
         // hash key to find target bucket index (bindex)
        int bindex = hash(key);
        MyLinkedList<KeyValuePair<K,V>> blist = this.table[bindex];
        KeyValuePair<K,V> kvp = new KeyValuePair(key, null); //no need for a value!
        /**
         * We can't simply use an iterator like we did for the put() method.
         * We can use indexOf -- capital O -- to find the index in the list,
         * if there is a matching element. In this case, a key.
         * The list is made of KeyValuePairs, - create one, and use equals()
         * well, indexOf will use our KeyValuePairs equals method, to find the
         * matching key if it exists.
         */
        int index = blist.indexOf(kvp); 
        if (index == -1) 
           return null;

        KeyValuePair<K,V> result = blist.remove(index); 
        this.nentries--;  // don't forget!
        return result.value;         
    }

    /**
     * Returns the number of key value pairs in the map.
     * @return number of key value pairs
     */
    public int size() {
       
        return nentries;
    }

    /**
     * Returns the map's current load factor, which is defined as the ratio between
     * the number of map entries and the number of buckets.
     * It is package private so we can use it in our unit test class.
     * 
     * @ return the hash map's current load factor.
     */
    double loadFactor(){
        return ((double) this.nentries) / this.nbuckets;
    }
    //
    // try it out!
    //
    // public static void main(String[] args) {
        
        // MyHashMap<String,String> hashy = new MyHashMap<String,String>();
        // hashy.put("authority", "supervisor"); 
        // hashy.remove("authority"); 
        
    // } 
}
